import 'dart:convert';

import 'package:analyzer/dart/element/element.dart';
import 'package:meta/meta.dart';
import 'package:dartrofit/dartrofit.dart';

/// Check parameter type which annotated on [QueryMap] whether is [Map]
void checkQueryMap(ParameterElement element) {
  if (!element.type.isDartCoreMap) {
    throw UnsupportedError(_errMsgBuilder(
        summary: 'The parameter type error which annotated on @QueryMap',
        details:
        'only support Map, but found ${element.type.getDisplayString()}',
        location: element.location));
  }
}

/// Check parameter type which annotated on [Body] whether is [String],
/// a [List<int>] or a [Map<String, String>]
void checkRequestBody(ParameterElement element) {
  final type = element.type;
  if (!type.isDartCoreString &&
      !(type.isDartCoreList && type.getDisplayString() == 'List<int>') &&
      !(type.isDartCoreMap &&
          type.getDisplayString() == 'Map<String, String>')) {
    throw UnsupportedError(_errMsgBuilder(
        summary: 'The parameter type error which annotated on @Body',
        details: 'only support String, List<int> or Map<String, String>, '
            'but found ${element.type.getDisplayString()}',
        location: element.location));
  }
}

/// Check and gets an [Encoding] object from the name of the character set
/// name, If character set is not supported error is raised.
Encoding checkEncodingName(ParameterElement element, String name) {
  final encoding = Encoding.getByName(name);
  if (encoding == null) {
    throw UnsupportedError(_errMsgBuilder(
        summary: 'Illegal encode name [$name] in Body',
        details:
        'see http://www.iana.org/assignments/character-sets/character-sets.xml.',
        location: element.location));
  }
  return encoding;
}

/// Build Parameter Error
Error parameterError(MethodElement element, String details) {
  return UnsupportedError(_errMsgBuilder(
      summary: 'Parameter Error',
      details: details,
      location: element.location));
}

/// Build Return Error
Error returnError(MethodElement element, String details) {
  return UnsupportedError(_errMsgBuilder(
      summary: 'Return Type Error',
      details: details,
      location: element.location));
}

/// Build Constructor Error
Error constructorError(ConstructorElement element, String details) {
  return UnsupportedError(_errMsgBuilder(
      summary: 'Constructor Error',
      details: details,
      location: element.location));
}

/// Build Class Error
Error classError(Element element, String details) {
  return UnsupportedError(_errMsgBuilder(
      summary: 'Class Error', details: details, location: element.location));
}

/// build form error message
String _errMsgBuilder(
    {@required summary, @required details, @required location}) {
  return '''$summary
  Details : $details
  Location: $location
  ''';
}