import 'package:analyzer/dart/element/element.dart';
import 'package:analyzer/dart/element/type.dart';
import 'package:analyzer/dart/element/visitor.dart';
import 'package:code_builder/code_builder.dart';
import 'package:dartrofit/dartrofit.dart';
import 'package:dartrofit_generator/src/checkers.dart';
import 'package:dartrofit_generator/src/dart_types.dart';
import 'package:dartrofit_generator/src/element_annotations.dart';
import 'package:dartrofit_generator/src/visitors/parameter_visitor.dart';

class ClassVisitor extends SimpleElementVisitor {
  Set<Method> methods;

  ClassVisitor();

  @override
  Object visitMethodElement(MethodElement element) {
    if (element.isAbstract) {
      final parameterVisitor = ParameterVisitor()
        ..requiredParameters = <Parameter>{}
        ..optionalParameters = <Parameter>{};
      element.visitChildren(parameterVisitor);

      final method = Method((b) => b
        ..returns = refer(element.returnType.getDisplayString())
        ..name = element.name
        ..requiredParameters.addAll(parameterVisitor.requiredParameters)
        ..optionalParameters.addAll(parameterVisitor.optionalParameters)
        ..body = Code(_bodyCode(element, parameterVisitor))
        ..annotations.addAll([const CodeExpression(Code('override'))]));

      methods.add(method);
    }
    return super.visitMethodElement(element);
  }

  String _bodyCode(MethodElement element, ParameterVisitor visitor) {
    String typeValueCodes;
    String returnTypeFullName;
    String dataTypeFullName;
    final returnType = element.returnType;
    // e.g: Subject<Response<Map<String, dynamic>>>
    returnTypeFullName = returnType.getDisplayString();
    if (returnType is ParameterizedType) {
      // check
      final rawReturnType = ParameterizedTypeValue(
          DartTypes.getLibraryName(returnType),
          returnType.element.name,
          returnType.getDisplayString(withNullability: false));

      final buffer = StringBuffer();
      buffer.writeln(
          "final rawReturnType = ParameterizedTypeValue('${DartTypes.getLibraryName(returnType)}', '${returnType.element.name}', '${returnType.getDisplayString(withNullability: false)}');");

      final innerType = _getParameterUpperBound(0, returnType);
      if (DartTypes.isDartrofitResponse(innerType)) {
        //R<Response<>>
        final responseType = innerType as ParameterizedType;
        final responseTypeValue = ParameterizedTypeValue(
            DartTypes.getLibraryName(responseType),
            responseType.element.name,
            responseType.getDisplayString(withNullability: false));
        rawReturnType.upperBoundAtIndex0 = responseTypeValue;

        final dataType = _getParameterUpperBound(0, responseType);
        responseTypeValue.upperBoundAtIndex0 = TypeValue(
            DartTypes.getLibraryName(dataType),
            dataType.element.name,
            dataType.element.getDisplayString(withNullability: false));
        dataTypeFullName = dataType.getDisplayString();

        buffer.writeln(
            "final responseTypeValue = ParameterizedTypeValue('${DartTypes.getLibraryName(responseType)}', '${responseType.element.name}', '${responseType.getDisplayString(withNullability: false)}');");
        buffer.writeln('rawReturnType.upperBoundAtIndex0 = responseTypeValue;');
        buffer.writeln(
            "responseTypeValue.upperBoundAtIndex0 = TypeValue('${DartTypes.getLibraryName(dataType)}', '${dataType.element.name}', '${dataType.getDisplayString(withNullability: false)}');");
      } else {
        //R<T>
        final dataType = TypeValue(
            DartTypes.getLibraryName(innerType),
            innerType.element.name,
            innerType.getDisplayString(withNullability: false));
        rawReturnType.upperBoundAtIndex0 = dataType;
        dataTypeFullName = innerType.getDisplayString();

        buffer.writeln(
            "final dataType = TypeValue('${DartTypes.getLibraryName(innerType)}', '${innerType.element.name}', '${innerType.getDisplayString(withNullability: false)}');");
        buffer.writeln('rawReturnType.upperBoundAtIndex0 = dataType;');
      }
      typeValueCodes = buffer.toString();
    } else {
      throw returnError(element, 'expect parameterized type');
    }

    // headers
    var headersInstanceStr = '';
    final headersHandlerStr = 'HeadersHandler()';
    final headersAnnotation = element.metadata
        .firstWhere(ElementAnnotations.isDartrofitHeaders, orElse: () => null);
    final headersValue = headersAnnotation
        ?.computeConstantValue()
        ?.getField('headers')
        ?.toListValue()
        ?.map((dartObject) => "'${dartObject.toStringValue()}'");
    final headersValueStr = headersValue?.join(', ');
    if (!Strings.isNullOrEmpty(headersValueStr)) {
      headersInstanceStr = '[$headersValueStr]';
    }

    // Multipart
    final isMultipart =
        element.metadata.any(ElementAnnotations.isDartrofitMultipart);

    String code;
    for (var annotation in element.metadata) {
      if (ElementAnnotations.isDartrofitGet(annotation)) {
        final relativeUrl = annotation
            .computeConstantValue()
            ?.getField('relativeUrl')
            ?.toStringValue();

        final handlersAndParameterNames =
            _parameterHandlersAndMethodParameterNamesByQueries(visitor);

        final parameterHandlersStringBuffer =
            StringBuffer(handlersAndParameterNames[0]);
        final methodParameterNamesStringBuffer =
            StringBuffer(handlersAndParameterNames[1]);

        if (!Strings.isNullOrEmpty(headersInstanceStr)) {
          parameterHandlersStringBuffer.write(', $headersHandlerStr');
          methodParameterNamesStringBuffer.write(', $headersInstanceStr');
        }

        _append(visitor.headerInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);
        _append(visitor.headerMapInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);

        _append(visitor.partFieldInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);
        _append(visitor.partFieldMapInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);
        _append(visitor.partFileInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);
        _append(visitor.partFileListInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);

        code = '''
        $typeValueCodes
        
        final methodInfo = MethodInfo('GET', '$relativeUrl', $isMultipart, [$parameterHandlersStringBuffer], rawReturnType);
        
        final invokes = [$methodParameterNamesStringBuffer];
          
        return dartrofit.parseAnnotation<$dataTypeFullName, $returnTypeFullName>(methodInfo).invoke(invokes);
        ''';
      } else if (ElementAnnotations.isDartrofitPost(annotation)) {
        final relativeUrl = annotation
            .computeConstantValue()
            ?.getField('relativeUrl')
            ?.toStringValue();

        final handlersAndParameterNames =
            _parameterHandlersAndMethodParameterNamesByQueriesAndBody(visitor);

        final parameterHandlersStringBuffer =
            StringBuffer(handlersAndParameterNames[0]);
        final methodParameterNamesStringBuffer =
            StringBuffer(handlersAndParameterNames[1]);

        if (!Strings.isNullOrEmpty(headersInstanceStr)) {
          parameterHandlersStringBuffer.write(', $headersHandlerStr');
          methodParameterNamesStringBuffer.write(', $headersInstanceStr');
        }

        _append(visitor.headerInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);
        _append(visitor.headerMapInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);

        _append(visitor.partFieldInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);
        _append(visitor.partFieldMapInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);
        _append(visitor.partFileInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);
        _append(visitor.partFileListInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);

        code = '''
      
        $typeValueCodes
        
        final methodInfo = MethodInfo('POST', '$relativeUrl', $isMultipart, [$parameterHandlersStringBuffer], rawReturnType);
        
        final invokes = [$methodParameterNamesStringBuffer];
        
        return dartrofit.parseAnnotation<$dataTypeFullName, $returnTypeFullName>(methodInfo).invoke(invokes);
        ''';
      } else if (ElementAnnotations.isDartrofitPut(annotation)) {
        final relativeUrl = annotation
            .computeConstantValue()
            ?.getField('relativeUrl')
            ?.toStringValue();

        final handlersAndParameterNames =
            _parameterHandlersAndMethodParameterNamesByQueriesAndBody(visitor);
        final parameterHandlersStringBuffer =
            StringBuffer(handlersAndParameterNames[0]);
        final methodParameterNamesStringBuffer =
            StringBuffer(handlersAndParameterNames[1]);

        if (!Strings.isNullOrEmpty(headersInstanceStr)) {
          parameterHandlersStringBuffer.write(', $headersHandlerStr');
          methodParameterNamesStringBuffer.write(', $headersInstanceStr');
        }

        _append(visitor.headerInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);
        _append(visitor.headerMapInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);

        _append(visitor.partFieldInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);
        _append(visitor.partFieldMapInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);
        _append(visitor.partFileInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);
        _append(visitor.partFileListInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);

        code = '''
      
        $typeValueCodes
        
        final methodInfo = MethodInfo('PUT', '$relativeUrl', $isMultipart, [$parameterHandlersStringBuffer], rawReturnType);
        
        final invokes = [$methodParameterNamesStringBuffer];
        
        return dartrofit.parseAnnotation<$dataTypeFullName, $returnTypeFullName>(methodInfo).invoke(invokes);
        ''';
      } else if (ElementAnnotations.isDartrofitPatch(annotation)) {
        final relativeUrl = annotation
            .computeConstantValue()
            ?.getField('relativeUrl')
            ?.toStringValue();

        final handlersAndParameterNames =
            _parameterHandlersAndMethodParameterNamesByQueriesAndBody(visitor);
        final parameterHandlersStringBuffer =
            StringBuffer(handlersAndParameterNames[0]);
        final methodParameterNamesStringBuffer =
            StringBuffer(handlersAndParameterNames[1]);

        if (!Strings.isNullOrEmpty(headersInstanceStr)) {
          parameterHandlersStringBuffer.write(', $headersHandlerStr');
          methodParameterNamesStringBuffer.write(', $headersInstanceStr');
        }

        _append(visitor.headerInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);
        _append(visitor.headerMapInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);

        _append(visitor.partFieldInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);
        _append(visitor.partFieldMapInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);
        _append(visitor.partFileInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);
        _append(visitor.partFileListInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);

        code = '''
      
        $typeValueCodes
        
        final methodInfo = MethodInfo('PATCH', '$relativeUrl', $isMultipart, [$parameterHandlersStringBuffer], rawReturnType);
        
        final invokes = [$methodParameterNamesStringBuffer];
        
        return dartrofit.parseAnnotation<$dataTypeFullName, $returnTypeFullName>(methodInfo).invoke(invokes);
        ''';
      } else if (ElementAnnotations.isDartrofitDelete(annotation)) {
        final relativeUrl = annotation
            .computeConstantValue()
            ?.getField('relativeUrl')
            ?.toStringValue();

        final handlersAndParameterNames =
            _parameterHandlersAndMethodParameterNamesByQueries(visitor);

        final parameterHandlersStringBuffer =
            StringBuffer(handlersAndParameterNames[0]);
        final methodParameterNamesStringBuffer =
            StringBuffer(handlersAndParameterNames[1]);

        if (!Strings.isNullOrEmpty(headersInstanceStr)) {
          parameterHandlersStringBuffer.write(', $headersHandlerStr');
          methodParameterNamesStringBuffer.write(', $headersInstanceStr');
        }

        _append(visitor.headerInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);
        _append(visitor.headerMapInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);

        _append(visitor.partFieldInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);
        _append(visitor.partFieldMapInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);
        _append(visitor.partFileInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);
        _append(visitor.partFileListInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);

        code = '''
      
        $typeValueCodes
        
        final methodInfo = MethodInfo('DELETE', '$relativeUrl', $isMultipart, [$parameterHandlersStringBuffer], rawReturnType);
        
        final invokes = [$methodParameterNamesStringBuffer];
        
        return dartrofit.parseAnnotation<$dataTypeFullName, $returnTypeFullName>(methodInfo).invoke(invokes);
        ''';
      } else if (ElementAnnotations.isDartrofitHead(annotation)) {
        final relativeUrl = annotation
            .computeConstantValue()
            ?.getField('relativeUrl')
            ?.toStringValue();

        final handlersAndParameterNames =
            _parameterHandlersAndMethodParameterNamesByQueries(visitor);

        final parameterHandlersStringBuffer =
            StringBuffer(handlersAndParameterNames[0]);
        final methodParameterNamesStringBuffer =
            StringBuffer(handlersAndParameterNames[1]);

        if (!Strings.isNullOrEmpty(headersInstanceStr)) {
          parameterHandlersStringBuffer.write(', $headersHandlerStr');
          methodParameterNamesStringBuffer.write(', $headersInstanceStr');
        }

        _append(visitor.headerInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);
        _append(visitor.headerMapInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);

        _append(visitor.partFieldInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);
        _append(visitor.partFieldMapInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);
        _append(visitor.partFileInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);
        _append(visitor.partFileListInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);

        code = '''
      
        $typeValueCodes
        
        final methodInfo = MethodInfo('HEAD', '$relativeUrl', $isMultipart, [$parameterHandlersStringBuffer], rawReturnType);
        
        final invokes = [$methodParameterNamesStringBuffer];
        
        return dartrofit.parseAnnotation<$dataTypeFullName, $returnTypeFullName>(methodInfo).invoke(invokes);
        ''';
      } else if (ElementAnnotations.isDartrofitOptions(annotation)) {
        final relativeUrl = annotation
            .computeConstantValue()
            ?.getField('relativeUrl')
            ?.toStringValue();

        final handlersAndParameterNames =
            _parameterHandlersAndMethodParameterNamesByQueries(visitor);

        final parameterHandlersStringBuffer =
            StringBuffer(handlersAndParameterNames[0]);
        final methodParameterNamesStringBuffer =
            StringBuffer(handlersAndParameterNames[1]);

        if (!Strings.isNullOrEmpty(headersInstanceStr)) {
          parameterHandlersStringBuffer.write(', $headersHandlerStr');
          methodParameterNamesStringBuffer.write(', $headersInstanceStr');
        }

        _append(visitor.headerInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);
        _append(visitor.headerMapInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);

        _append(visitor.partFieldInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);
        _append(visitor.partFieldMapInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);
        _append(visitor.partFileInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);
        _append(visitor.partFileListInfo, parameterHandlersStringBuffer,
            methodParameterNamesStringBuffer);

        code = '''
      
        $typeValueCodes
        
        final methodInfo = MethodInfo('OPTIONS', '$relativeUrl', $isMultipart, [$parameterHandlersStringBuffer], rawReturnType);
        
        final invokes = [$methodParameterNamesStringBuffer];
        
        return dartrofit.parseAnnotation<$dataTypeFullName, $returnTypeFullName>(methodInfo).invoke(invokes);
        ''';
      }
    }
    return code;
  }

  static List<String> _parameterHandlersAndMethodParameterNamesByQueriesAndBody(
      ParameterVisitor visitor) {
    var parameterHandlers = '';
    var methodParameterNames = '';

    final handlersAndParameterNames =
        _parameterHandlersAndMethodParameterNamesByQueries(visitor);
    final queryParameterHandlerInstanceStr = handlersAndParameterNames[0];
    final queryMethodParameterName = handlersAndParameterNames[1];

    final bodyInfo = visitor.bodyInfo;
    final bodyParameterHandlerInstanceStr =
        bodyInfo?.parameterHandlerInstanceStr;
    final bodyMethodParameterName = bodyInfo?.methodParameterName;

    if (Strings.isNullOrEmpty(queryParameterHandlerInstanceStr)) {
      if (!Strings.isNullOrEmpty(bodyParameterHandlerInstanceStr)) {
        parameterHandlers = bodyParameterHandlerInstanceStr;
        methodParameterNames = bodyMethodParameterName;
      }
    } else {
      if (Strings.isNullOrEmpty(bodyParameterHandlerInstanceStr)) {
        parameterHandlers = queryParameterHandlerInstanceStr;
        methodParameterNames = queryMethodParameterName;
      } else {
        parameterHandlers =
            '$queryParameterHandlerInstanceStr, $bodyParameterHandlerInstanceStr';
        methodParameterNames =
            '$queryMethodParameterName, $bodyMethodParameterName';
      }
    }
    return <String>[parameterHandlers, methodParameterNames];
  }

  static List<String> _parameterHandlersAndMethodParameterNamesByQueries(
      ParameterVisitor visitor) {
    final queryInfos = visitor.queryInfos;
    final queryMapInfos = visitor.queryMapInfos;

    if (queryInfos.isEmpty && queryMapInfos.isEmpty) {
      return <String>['', ''];
    }
    final queryInfoHandlers = queryInfos
        .map((queryInfo) => queryInfo.parameterHandlerInstanceStr)
        .join(', ');
    final queryMapInfoHandlers = queryMapInfos
        .map((queryMapInfo) => queryMapInfo.parameterHandlerInstanceStr)
        .join(', ');

    final queryInfoParameterNames =
        queryInfos.map((queryInfo) => queryInfo.methodParameterName).join(', ');
    final queryMapInfoParameterNames = queryMapInfos
        .map((queryMapInfo) => queryMapInfo.methodParameterName)
        .join(', ');

    var handlers = '';
    var parameterNames = '';
    if (queryInfos.isNotEmpty) {
      handlers = queryInfoHandlers;
      parameterNames = queryInfoParameterNames;
      if (queryMapInfos.isNotEmpty) {
        handlers += ', $queryMapInfoHandlers';
        parameterNames += ', $queryMapInfoParameterNames';
      }
    } else {
      handlers = queryMapInfoHandlers;
      parameterNames = queryMapInfoParameterNames;
    }
    return <String>[handlers, parameterNames];
  }

  static void _append(
      AnnotationInfo info,
      StringBuffer parameterHandlersStringBuffer,
      StringBuffer methodParameterNamesStringBuffer) {
    if (info != null) {
      if (parameterHandlersStringBuffer.isEmpty) {
        parameterHandlersStringBuffer.write(info.parameterHandlerInstanceStr);
      } else {
        parameterHandlersStringBuffer
            .write(', ${info.parameterHandlerInstanceStr}');
      }

      if (methodParameterNamesStringBuffer.isEmpty) {
        methodParameterNamesStringBuffer.write(info.methodParameterName);
      } else {
        methodParameterNamesStringBuffer.write(', ${info.methodParameterName}');
      }
    }
  }

  static DartType _getParameterUpperBound(int index, ParameterizedType type) {
    final types = type.typeArguments;
    if (index < 0 || index >= types.length) {
      throw ArgumentError(
          'Index $index not in range [0, ${types.length}) for $type');
    }
    return types[index];
  }

  static bool _isNullOrEmpty<T>(Iterable<T> iterable) =>
      iterable == null || iterable.isEmpty;
}
